"""
Public namespace for ABCs corresponding to protocol classes.
"""

from ._abcs import (
    Ordered,
    Comparable,
    Additive,
    Multiplicative,
    Algebraic,
    Real,
    Scalar,
    Sequence,
    Array,
)


__all__ = [
    'Ordered',
    'Comparable',
    'Additive',
    'Multiplicative',
    'Algebraic',
    'Real',
    'Scalar',
    'Sequence',
    'Array',
]


